@@ -54,6 +54,7 @@ TODO:
 #include "mysql/psi/mysql_file.h"
 
 PSI_file_key key_file_io_cache;
+TMPFILE_SIZE_CB update_tmp_file_size= 0;
 
 #define lock_append_buffer(info) \
   mysql_mutex_lock(&(info)->append_buffer_lock)
@@ -73,6 +74,58 @@ int (*_my_b_encr_read)(IO_CACHE *info,uchar *Buffer,size_t Count)= 0;
 int (*_my_b_encr_write)(IO_CACHE *info,const uchar *Buffer,size_t Count)= 0;
 
 
+static inline my_bool tmp_file_track(IO_CACHE *info, ulonglong file_size)
+{
+  if ((info->myflags & (MY_TRACK | MY_TRACK_WITH_LIMIT)) &&
+      update_tmp_file_size)
+  {
+    if (info->tracking.file_size < file_size)
+    {
+      int error;
+      info->tracking.file_size= file_size;
+      if ((error= update_tmp_file_size(&info->tracking,
+                                       !(info->myflags &
+                                         MY_TRACK_WITH_LIMIT))))
+      {
+        if (info->myflags & MY_WME)
+          my_error(error, MYF(0));
+        info->error= -1;
+        return 1;
+      }
+    }
+  }
+  return 0;
+}
+
+my_bool io_cache_tmp_file_track(IO_CACHE *info, ulonglong file_size)
+{
+  return tmp_file_track(info, file_size);
+}
+
+
+/**
+  Reset tmp space tracking for the data in the io cache
+
+  This is called when deleting or truncating the
+  cached file.
+*/
+
+static void reset_tracking_io_cache(IO_CACHE *info)
+{
+  if ((info->myflags & (MY_TRACK | MY_TRACK_WITH_LIMIT)) &&
+      info->tracking.file_size)
+  {
+    info->tracking.file_size= 0;
+    update_tmp_file_size(&info->tracking, 1);
+  }
+}
+
+void truncate_io_cache(IO_CACHE *info)
+{
+  if (my_chsize(info->file, 0, 0, MYF(MY_WME)) == 0)
+    reset_tracking_io_cache(info);
+}
+
 
 static void
 init_functions(IO_CACHE* info)
@@ -153,7 +206,7 @@ init_functions(IO_CACHE* info)
 int init_io_cache_ext(IO_CACHE *info, File file, size_t cachesize,
                       enum cache_type type, my_off_t seek_offset,
                       pbool use_async_io, myf cache_myflags,
-                      PSI_file_key file_key)
+                      PSI_file_key file_key __attribute__((unused)))
 {
   size_t min_cache;
   my_off_t pos;
@@ -169,6 +222,8 @@ int init_io_cache_ext(IO_CACHE *info, File file, size_t cachesize,
   info->buffer=0;
   info->seek_not_done= 0;
   info->next_file_user= NULL;
+  info->tracking.previous_file_size= 0;
+  info->tracking.file_size= 0;
 
   if (file >= 0)
   {
@@ -407,11 +462,18 @@ void seek_io_cache(IO_CACHE *cache, my_off_t needed_offset)
 }
 
 
-/*
+/**
   Use this to reset cache to re-start reading or to change the type
   between READ_CACHE <-> WRITE_CACHE
   If we are doing a reinit of a cache where we have the start of the file
   in the cache, we are reusing this memory without flushing it to disk.
+
+  @param info          IO_CACHE
+  @param type          READ_CACHE or WRITE_CACHE
+  @param seek_offset   Where to start reading or writing
+  @param use_async_io  Not used
+  @param clear_cache   0  No clear, keep all information
+                       1  truncate file. seek_offset has to be 0.
 */
 
 my_bool reinit_io_cache(IO_CACHE *info, enum cache_type type,
@@ -478,8 +540,24 @@ my_bool reinit_io_cache(IO_CACHE *info, enum cache_type type,
       }
     }
     /* flush cache if we want to reuse it */
-    if (!clear_cache && my_b_flush_io_cache(info,1))
-      DBUG_RETURN(1);
+    if (!clear_cache && info->type == WRITE_CACHE)
+    {
+      int ret;
+      myf save_flags;
+      if (type == WRITE_CACHE && seek_offset > info->pos_in_file)
+      {
+        /* Write only up to where we will start next write */
+        my_off_t buffer_used= seek_offset - info->pos_in_file;
+        info->write_pos= info->write_buffer + buffer_used;
+      }
+      save_flags= info->myflags;
+      /* Allow temporary space over usage. Will be detected on next write */
+      info->myflags&= ~MY_TRACK_WITH_LIMIT;
+      ret= my_b_flush_io_cache(info,1);
+      info->myflags= save_flags;
+      if (ret)
+        DBUG_RETURN(ret);
+    }
     info->pos_in_file=seek_offset;
     /* Better to do always do a seek */
     info->seek_not_done=1;
@@ -1523,7 +1601,8 @@ int _my_b_cache_write(IO_CACHE *info, const uchar *Buffer, size_t Count)
     }
     info->seek_not_done=0;
   }
-  if (mysql_file_write(info->file, Buffer, Count, info->myflags | MY_NABP))
+  if (tmp_file_track(info, info->pos_in_file + Count) ||
+      mysql_file_write(info->file, Buffer, Count, info->myflags | MY_NABP))
     return info->error= -1;
 
   info->pos_in_file+= Count;
@@ -1592,7 +1671,8 @@ int my_b_append(IO_CACHE *info, const uchar *Buffer, size_t Count)
   if (Count >= IO_SIZE)
   {					/* Fill first intern buffer */
     length= IO_ROUND_DN(Count);
-    if (mysql_file_write(info->file,Buffer, length, info->myflags | MY_NABP))
+    if (tmp_file_track(info, info->end_of_file + length) ||
+        mysql_file_write(info->file,Buffer, length, info->myflags | MY_NABP))
     {
       unlock_append_buffer(info);
       return info->error= -1;
@@ -1710,15 +1790,18 @@ int my_b_flush_io_cache(IO_CACHE *info, int need_append_buffer_lock)
 
     if ((length=(size_t) (info->write_pos - info->write_buffer)))
     {
+      my_off_t eof= info->end_of_file + info->write_pos - info->append_read_pos;
       if (append_cache)
       {
-        if (mysql_file_write(info->file, info->write_buffer, length,
+        if (tmp_file_track(info, eof) ||
+            mysql_file_write(info->file, info->write_buffer, length,
                              info->myflags | MY_NABP))
         {
+          UNLOCK_APPEND_BUFFER;
           info->error= -1;
           DBUG_RETURN(-1);
         }
-        info->end_of_file+= info->write_pos - info->append_read_pos;
+        info->end_of_file= eof;
         info->append_read_pos= info->write_buffer;
         DBUG_ASSERT(info->end_of_file == mysql_file_tell(info->file, MYF(0)));
       }
@@ -1726,8 +1809,10 @@ int my_b_flush_io_cache(IO_CACHE *info, int need_append_buffer_lock)
       {
         int res= info->write_function(info, info->write_buffer, length);
         if (res)
+        {
+          UNLOCK_APPEND_BUFFER;
           DBUG_RETURN(res);
-
+        }
         set_if_bigger(info->end_of_file, info->pos_in_file);
       }
       info->write_end= (info->write_buffer + info->buffer_length -
@@ -1735,7 +1820,7 @@ int my_b_flush_io_cache(IO_CACHE *info, int need_append_buffer_lock)
       info->write_pos= info->write_buffer;
       ++info->disk_writes;
       UNLOCK_APPEND_BUFFER;
-      DBUG_RETURN(info->error);
+      DBUG_RETURN(0);
     }
   }
   UNLOCK_APPEND_BUFFER;
@@ -1785,14 +1870,15 @@ int end_io_cache(IO_CACHE *info)
     /* Destroy allocated mutex */
     mysql_mutex_destroy(&info->append_buffer_lock);
   }
+  reset_tracking_io_cache(info);
   info->share= 0;
-  info->type= TYPE_NOT_SET;                  /* Ensure that flush_io_cache() does nothing */
-  info->write_end= 0;                        /* Ensure that my_b_write() fails */
-  info->write_function= 0;                   /* my_b_write will crash if used */
+  info->type= TYPE_NOT_SET; /* Ensure that flush_io_cache() does nothing */
+  info->write_end= 0;       /* Ensure that my_b_write() fails */
+  info->write_function= 0;  /* my_b_write will crash if used */
+
   DBUG_RETURN(error);
 } /* end_io_cache */
 
-
 /**********************************************************************
  Testing of MF_IOCACHE
 **********************************************************************/
