@@ -78,7 +78,7 @@
 #define DEFAULT_AWS_HOST_NAME "s3.amazonaws.com"
 
 static PAGECACHE s3_pagecache;
-static ulong s3_block_size, s3_protocol_version;
+static ulong s3_block_size, s3_protocol_version, s3_provider;
 static ulong s3_pagecache_division_limit, s3_pagecache_age_threshold;
 static ulong s3_pagecache_file_hash_size;
 static ulonglong s3_pagecache_buffer_size;
@@ -86,6 +86,8 @@ static char *s3_bucket, *s3_access_key=0, *s3_secret_key=0, *s3_region;
 static char *s3_host_name;
 static int s3_port;
 static my_bool s3_use_http;
+static my_bool s3_ssl_no_verify;
+static my_bool s3_no_content_type;
 static char *s3_tmp_access_key=0, *s3_tmp_secret_key=0;
 static my_bool s3_debug= 0, s3_slave_ignore_updates= 0;
 static my_bool s3_replicate_alter_as_create_select= 0;
@@ -121,6 +123,29 @@ static void update_secret_key(MYSQL_THD thd,
   }
 }
 
+static void update_s3_debug(MYSQL_THD thd,
+                            struct st_mysql_sys_var *var
+                            __attribute__((unused)),
+                            void *var_ptr __attribute__((unused)),
+                            const void *save)
+{
+  char new_state= *(char *) save;
+  if (s3_debug != new_state)
+  {
+    s3_debug= new_state;
+    if (s3_hton)                                // If library is initalized
+    {
+      ms3_debug(new_state);
+      if (!new_state)
+      {
+        /* Ensure that all logging is written to log */
+        fflush(stderr);
+      }
+    }
+  }
+}
+
+
 /* Define system variables for S3 */
 
 static MYSQL_SYSVAR_ULONG(block_size, s3_block_size,
@@ -129,9 +154,9 @@ static MYSQL_SYSVAR_ULONG(block_size, s3_block_size,
        4*1024*1024, 65536, 16*1024*1024, 8192);
 
 static MYSQL_SYSVAR_BOOL(debug, s3_debug,
-       PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
+       PLUGIN_VAR_RQCMDARG,
       "Generates trace file from libmarias3 on stderr for debugging",
-       0, 0, 0);
+       0, update_s3_debug, 0);
 
 static MYSQL_SYSVAR_BOOL(slave_ignore_updates, s3_slave_ignore_updates,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
@@ -147,7 +172,10 @@ static MYSQL_SYSVAR_BOOL(replicate_alter_as_create_select,
 static MYSQL_SYSVAR_ENUM(protocol_version, s3_protocol_version,
                          PLUGIN_VAR_RQCMDARG,
                          "Protocol used to communication with S3. One of "
-                         "\"Auto\", \"Amazon\" or \"Original\".",
+                         "\"Auto\", \"Legacy\", \"Original\", \"Amazon\", "
+                         "\"Path\" or \"Domain\". "
+                         "Note: \"Legacy\", \"Original\" and \"Amazon\" are "
+                         "deprecated",
                          NULL, NULL, 0, &s3_protocol_typelib);
 
 static MYSQL_SYSVAR_ULONG(pagecache_age_threshold,
@@ -155,14 +183,14 @@ static MYSQL_SYSVAR_ULONG(pagecache_age_threshold,
        "This characterizes the number of hits a hot block has to be untouched "
        "until it is considered aged enough to be downgraded to a warm block. "
        "This specifies the percentage ratio of that number of hits to the "
-       "total number of blocks in the page cache.", 0, 0,
+       "total number of blocks in the page cache", 0, 0,
        300, 100, ~ (ulong) 0L, 100);
 
 static MYSQL_SYSVAR_ULONGLONG(pagecache_buffer_size, s3_pagecache_buffer_size,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
        "The size of the buffer used for index blocks for S3 tables. "
        "Increase this to get better index handling (for all reads and "
-       "multiple writes) to as much as you can afford.", 0, 0,
+       "multiple writes) to as much as you can afford", 0, 0,
         128*1024*1024, 1024*1024*32, ~(ulonglong) 0, 8192);
 
 static MYSQL_SYSVAR_ULONG(pagecache_division_limit,
@@ -177,7 +205,7 @@ static MYSQL_SYSVAR_ULONG(pagecache_file_hash_size,
        "Number of hash buckets for open files.  If you have a lot "
        "of S3 files open you should increase this for faster flush of "
        "changes. A good value is probably 1/10 of number of possible open "
-       "S3 files.", 0,0, 512, 32, 16384, 1);
+       "S3 files", 0,0, 512, 32, 16384, 1);
 
 static MYSQL_SYSVAR_STR(bucket, s3_bucket,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
@@ -196,6 +224,10 @@ static MYSQL_SYSVAR_BOOL(use_http, s3_use_http,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
       "If true, force use of HTTP protocol",
        NULL /*check*/, NULL /*update*/, 0 /*default*/);
+static MYSQL_SYSVAR_BOOL(ssl_no_verify, s3_ssl_no_verify,
+       PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
+      "If true, SSL certificate verification for the S3 endpoint is disabled",
+       NULL, NULL, 0);
 static MYSQL_SYSVAR_STR(access_key, s3_tmp_access_key,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY | PLUGIN_VAR_MEMALLOC,
       "AWS access key",
@@ -208,6 +240,15 @@ static MYSQL_SYSVAR_STR(region, s3_region,
        PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
       "AWS region",
        0, 0, "");
+static MYSQL_SYSVAR_BOOL(no_content_type, s3_no_content_type,
+       PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_READONLY,
+      "If true, disables the Content-Type header, required for some providers",
+      NULL, NULL, 0);
+static MYSQL_SYSVAR_ENUM(provider, s3_provider,
+                         PLUGIN_VAR_RQCMDARG,
+                         "Enable S3 provider specific compatibility tweaks "
+                         "\"Default\", \"Amazon\", or \"Huawei\"",
+                         NULL, NULL, 0, &s3_provider_typelib);
 
 ha_create_table_option s3_table_option_list[]=
 {
@@ -238,6 +279,7 @@ ha_s3::ha_s3(handlerton *hton, TABLE_SHARE *table_arg)
   /* Remove things that S3 doesn't support */
   int_table_flags&= ~(HA_BINLOG_ROW_CAPABLE | HA_BINLOG_STMT_CAPABLE |
                       HA_CAN_EXPORT);
+  int_table_flags|= HA_NO_ONLINE_ALTER;
   can_enable_indexes= 0;
 }
 
@@ -293,6 +335,9 @@ static my_bool s3_info_init(S3_INFO *info)
   lex_string_set(&info->host_name,  s3_host_name);
   info->port= s3_port;
   info->use_http= s3_use_http;
+  info->ssl_no_verify= s3_ssl_no_verify;
+  info->no_content_type = s3_no_content_type;
+  info->provider= s3_provider;
   lex_string_set(&info->access_key, s3_access_key);
   lex_string_set(&info->secret_key, s3_secret_key);
   lex_string_set(&info->region,     s3_region);
@@ -738,7 +783,15 @@ static int s3_hton_panic(handlerton *hton, ha_panic_function flag)
 
 
 /**
-  Check if a table is in S3 as part of discovery
+  Check if a table is in S3 as part of discovery. Returns TABLE_SHARE if found.
+
+  @param hton           S3 handlerton
+  @param thd            MariaDB thd
+  @param [out] share    If table exists, this is updated to contain the found
+                        TABLE_SHARE (based on the .frm in S3)
+
+  @return 0  Table exists
+  @return #  Error number
 */
 
 static int s3_discover_table(handlerton *hton, THD* thd, TABLE_SHARE *share)
@@ -783,13 +836,13 @@ static int s3_discover_table(handlerton *hton, THD* thd, TABLE_SHARE *share)
    @return 1 frm exists
 */
 
-static int s3_discover_table_existance(handlerton *hton, const char *db,
+static int s3_discover_table_existence(handlerton *hton, const char *db,
                                        const char *table_name)
 {
   S3_INFO s3_info;
   ms3_st *s3_client;
   int res;
-  DBUG_ENTER("s3_discover_table_existance");
+  DBUG_ENTER("s3_discover_table_existence");
 
   /* Ignore names in "mysql" database to speed up boot */
   if (!strcmp(db, MYSQL_SCHEMA_NAME.str))
@@ -819,7 +872,7 @@ static int s3_discover_table_existance(handlerton *hton, const char *db,
 */
 
 static int s3_discover_table_names(handlerton *hton __attribute__((unused)),
-                                   LEX_CSTRING *db,
+                                   const LEX_CSTRING *db,
                                    MY_DIR *dir __attribute__((unused)),
                                    handlerton::discovered_list *result)
 {
@@ -1013,13 +1066,13 @@ static int ha_s3_init(void *p)
   s3_hton->table_options= s3_table_option_list;
   s3_hton->discover_table= s3_discover_table;
   s3_hton->discover_table_names= s3_discover_table_names;
-  s3_hton->discover_table_existence= s3_discover_table_existance;
+  s3_hton->discover_table_existence= s3_discover_table_existence;
   s3_hton->notify_tabledef_changed= s3_notify_tabledef_changed;
   s3_hton->create_partitioning_metadata= s3_create_partitioning_metadata;
   s3_hton->tablefile_extensions= no_exts;
   s3_hton->commit= 0;
   s3_hton->rollback= 0;
-  s3_hton->checkpoint_state= 0;
+  s3_hton->disable_internal_writes= 0;
   s3_hton->flush_logs= 0;
   s3_hton->show_status= 0;
   s3_hton->prepare_for_backup= 0;
@@ -1041,7 +1094,7 @@ static int ha_s3_init(void *p)
   s3_pagecache.big_block_free= s3_free;
   s3_init_library();
   if (s3_debug)
-    ms3_debug();
+    ms3_debug(1);
 
   struct s3_func s3f_real =
   {
@@ -1086,12 +1139,15 @@ static struct st_mysql_sys_var* system_variables[]= {
   MYSQL_SYSVAR(host_name),
   MYSQL_SYSVAR(port),
   MYSQL_SYSVAR(use_http),
+  MYSQL_SYSVAR(ssl_no_verify),
   MYSQL_SYSVAR(bucket),
   MYSQL_SYSVAR(access_key),
   MYSQL_SYSVAR(secret_key),
   MYSQL_SYSVAR(region),
   MYSQL_SYSVAR(slave_ignore_updates),
   MYSQL_SYSVAR(replicate_alter_as_create_select),
+  MYSQL_SYSVAR(no_content_type),
+  MYSQL_SYSVAR(provider),
   NULL
 };
 
